unit DUnitAssert;
// Assert extension for DUnit
// (c) 2008  Tiriss - http://www.tiriss.com
//
// VERSION: V1.2
//
// PURPOSE
// Assert for DUnit allows to use the (NUnit) Assert.AreEqual etc. syntax to be used in Delphi's DUnit
// It was created to
// * make switching between (C#).Net and Delphi easier
// * be able to only learn one syntax.
//
// HOME
// http://www.tiriss.com/AssertForDUnit
//
// LICENSE
// This sourcecode is free software. It comes without any warranty, to the extent permitted by applicable law.
// You can redistribute it and/or modify it under the term that you keep a clear reference to its original home
// (http://www.tiriss.com/AssertForDUnit)
//
// ADDITIONS
// This is still a work in progress (like most software), however the TDUnitAssert class should have most methods.
// If you have additions, and would like them to be considered to be added, then please send them to tiriss@tiriss.com

interface

uses
  TestFramework, TypInfo, SysUtils;

//compatibility with previous Delphi versions
// todo check Delphi versions other then 5,7 & 2007
{$IFDEF CONDITIONALEXPRESSIONS}
  {$IF COMPILERVERSION >= 17}
    {$DEFINE USEINLINE}
    {$DEFINE DEPRECATED}
  {$IFEND}
{$ENDIF}

{$IFDEF VER130} // Delphi 5 doesn't have IInterface yet
type
  IInterface = IUnknown;
{$ELSE}
  {$IFNDEF VER140}
    //{$DEFINE DEPRECATED} // deprecated is in D7 and up, but even gives compile warning on implementation!
  {$ENDIF}
{$ENDIF}

type
  TMaxSet = set of 0..255;

  TAssertTestCase = class; // forward decl
  /// <summary>
  /// TDUnitAssert is a Compatibility class to be able to use Assert (like C#) instead of Check
  /// By defining string as const and using inline we have (almost) no overhead for this!
  /// For example: overhead is 1 instruction/4 bytes for the IsTrue case
    /// </summary>
  TDUnitAssert = class
  private
    FTestCase: TAssertTestCase;
    class function GetSetElementTypeData(aSetTypeInfo: PTypeInfo): PTypeData;
    class procedure ConvertSetToTMaxSet(const aSet; aElementTypeData: PTypeData; out aMaxSet:TMaxSet); {$IFDEF USEINLINE}inline;{$ENDIF}
  protected
    class function SetToString(const aSetValue; aSetTypeInfo: PTypeInfo): string;
    procedure AreEqualDate(const aExpected, aActual: TDateTime; const aMessage: string = ''); overload;
    procedure AreEqualEnum(aExpectedEnum, aActualEnum: LongWord; aEnumTypeInfo: PTypeInfo; aMsg: string = ''); overload;
    procedure AreNotEqualEnum(aNotExpectedEnum, aActualEnum: LongWord; aEnumTypeInfo: PTypeInfo; aMsg: string = '');
        overload;
    procedure AreEqualSet(const aExpectedSet, aActualSet; aSetTypeInfo: PTypeInfo; aMsg: string = ''); overload;
    procedure AreNotEqualDate(const aNotExpected, aActual: TDateTime; const aMessage: string = ''); overload;
    procedure AreNotEqualSet(const aNotExpectedSet, aActualSet; aSetTypeInfo: PTypeInfo; aMsg: string = ''); overload;
    property TestCase: TAssertTestCase read FTestCase;
  public
    constructor Create(aTestCase: TAssertTestCase);
    procedure AreEqual(aExpected, aActual: Boolean; const aMessage: string = ''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreEqual(aExpected, aActual : Extended; delta : Extended; const aMessage : string = ''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreEqual(aExpected, aActual: Extended; const aMessage: string = ''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreEqual(aExpected, aActual: Integer; const aMessage: string = ''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreEqual(const aExpected, aActual: string; const aMessage: string = ''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreEqual(aExpected, aActual: TClass; const aMessage: string = ''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreEqual(aExpected, aActual: WideString; const aMessage: string = ''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreEqual(const aExpected, aActual: TDateTime; const aMessage: string = ''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreEqual(const aExpectedValue, aActualValue; aTypeInfo: PTypeInfo; aMsg: string = ''); overload;
    /// <summary>
    /// AreEqualString is for earlier Delphi versions that give "Ambiguous overloaded call to 'CheckEquals'"
    /// </summary>
    procedure AreEqualString(const aExpected, aActual: string; const aMessage: string = ''); overload; {$IFDEF DEPRECATED}deprecated;{$ENDIF}{$IFDEF USEINLINE}inline;{$ENDIF}
    /// <summary>
    /// AreEqualWideString is for earlier Delphi versions that give "Ambiguous overloaded call to 'CheckEquals'"
    /// </summary>
    procedure AreEqualWideString(const aExpected, aActual: WideString; const aMessage: string = ''); overload; {$IFDEF DEPRECATED}deprecated;{$ENDIF}{$IFDEF USEINLINE}inline;{$ENDIF}
    /// <summary>
    /// We don't use an AreEqual overload for AreEqualVariant, because AreEqualVariant will match all types
    /// </summary>
    procedure AreEqualVariant(const aExpected, aActual: Variant; aMsg: string = ''); overload;
    procedure AreSame(aExpected, aActual: IInterface; const aMessage: string = ''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreSame(aExpected, aActual: TObject; const aMessage: string = ''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure Fail(const aMessage: string); {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure IsFalse(aCondition: Boolean; const aMessage: string=''); {$IFDEF USEINLINE}inline;{$ENDIF}
    // mstest seems to have the parameters the otherway around; this is the NUnit, MbUnit syntax
    procedure IsInstanceOfType(const aExpectedClass: TClass; const aObject: TObject; const aMessage: string = '');
        overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure IsNotNull(aObj: TObject; const aMessage: string=''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure IsNotNull(aObj: IInterface; const aMessage: string=''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure IsNull(aObj: IInterface; const aMessage: string=''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure IsNull(aObj: TObject; const aMessage: string=''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    // mstest seems to have the parameters the otherway around; this is the NUnit, MbUnit syntax
    procedure IsNotInstanceOfType(const aExpectedClass: TClass; const aObject: TObject; const aMessage: string = '');
        overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreNotSame(aNotExpected, aActual: TObject; const aMessage: string = ''); overload;{$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreNotSame(aNotExpected, aActual: IInterface; const aMessage: string = ''); overload;{$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreNotEqual(aNotExpected, aActual: Integer; const aMessage: string = ''); overload;{$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreNotEqual(aNotExpected, aActual: Extended; const aMessage: string = ''); overload;{$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreNotEqual(aNotExpected, aActual: Extended; delta: Extended; const aMessage: string = ''); overload;{$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreNotEqual(aNotExpected, aActual: Boolean; const aMessage: string = ''); overload;{$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreNotEqual(aNotExpected, aActual: TClass; const aMessage: string = ''); overload;
    procedure AreNotEqual(const aNotExpected, aActual: string; const aMessage: string = ''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure AreNotEqual(const aNotExpected, aActual: WideString; const aMessage: string = ''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    /// <summary>
    /// AreNotEqualString is for earlier Delphi versions that give "Ambiguous overloaded call to 'CheckEquals'"
    /// </summary>
    procedure AreNotEqualString(const aNotExpected, aActual: string; const aMessage: string = ''); overload; {$IFDEF DEPRECATED}deprecated;{$ENDIF}{$IFDEF USEINLINE}inline;{$ENDIF}
    /// <summary>
    /// AreNotEqualWideString is for earlier Delphi versions that give "Ambiguous overloaded call to 'CheckEquals'"
    /// </summary>
    procedure AreNotEqualWideString(const aNotExpected, aActual: WideString; const aMessage: string = ''); overload; {$IFDEF DEPRECATED}deprecated;{$ENDIF}{$IFDEF USEINLINE}inline;{$ENDIF}
    /// <summary>
    /// We don't use an AreNotEqual overload for AreNotEqualVariant, because AreNotEqualVariant will match all types and that is probably  not what you want
    /// </summary>
    procedure AreNotEqualVariant(const aNotExpected, aActual: Variant; aMsg: string = ''); overload;
    procedure AreNotEqual(const aExpectedValue, aActualValue; aTypeInfo: PTypeInfo; aMsg: string = ''); overload;
    procedure AreNotEqual(const aNotExpected, aActual: TDateTime; const aMessage: string = ''); overload; {$IFDEF USEINLINE}inline;{$ENDIF}
    procedure IsTrue(aCondition: Boolean; const aMessage: string=''); {$IFDEF USEINLINE}inline;{$ENDIF}
  end;

  {$M+}
  /// <summary>
  /// Compatibility class to be able to use Assert instead of Check
  /// inherit your test class from this to be able to use the Assert syntax
  /// </summary>
  TAssertTestCase = class(TTestCase)
  private
    FAssert: TDUnitAssert;
  protected
    /// <summary>
    /// CreateAssert allows descendant to create their own (extended) TDUnitAssert class
    /// </summary>
    /// <returns> TDUnitAssert
    /// </returns>
    function CreateAssert: TDUnitAssert; virtual;
    procedure AssertCalled; {$IFDEF USEINLINE}inline;{$ENDIF}
  public
    constructor Create(MethodName: string); override;
    destructor Destroy; override;
    property Assert: TDUnitAssert read FAssert;
  end;
  {$M-}

implementation

{$IFNDEF VER130}
uses Variants;
{$ENDIF}

//uses SysUtils;

{ TDUnitAssert }

constructor TDUnitAssert.Create(aTestCase: TAssertTestCase);
begin
  FTestCase := aTestCase;
end;

procedure TDUnitAssert.AreEqual(aExpected, aActual: Boolean; const aMessage: string = '');
begin
  FTestCase.CheckEquals(aExpected, aActual, aMessage);
end;

procedure TDUnitAssert.AreEqual(aExpected, aActual : Extended; delta : Extended; const aMessage : string = '');
begin
  FTestCase.CheckEquals(aExpected, aActual, delta, aMessage);
end;

procedure TDUnitAssert.AreEqual(aExpected, aActual: Extended; const aMessage: string = '');
begin
  FTestCase.CheckEquals(aExpected, aActual, aMessage);
end;

procedure TDUnitAssert.AreEqual(aExpected, aActual: Integer; const aMessage: string = '');
begin
  FTestCase.CheckEquals(aExpected, aActual, aMessage);
end;

procedure TDUnitAssert.AreEqual(const aExpected, aActual: string; const aMessage: string = '');
begin
  {$IFNDEF VER130}
  FTestCase.CheckEquals(aExpected, aActual, aMessage);
  {$ELSE} // D5
  AreEqualString(aExpected, aActual, aMessage);
  {$ENDIF}
end;

procedure TDUnitAssert.AreEqual(aExpected, aActual: TClass; const aMessage: string = '');
begin
  FTestCase.CheckEquals(aExpected, aActual, aMessage);
end;

procedure TDUnitAssert.AreEqual(aExpected, aActual: WideString; const aMessage: string = '');
begin
  {$IFNDEF VER130}
  FTestCase.CheckEquals(aExpected, aActual, aMessage);
  {$ELSE} // D5
  AreEqualWideString(aExpected, aActual, aMessage);
  {$ENDIF}
end;

procedure TDUnitAssert.AreEqual(const aExpected, aActual: TDateTime; const aMessage: string = '');
begin
  AreEqualDate(aExpected, aActual, aMessage);
end;

procedure TDUnitAssert.AreEqual(const aExpectedValue, aActualValue; aTypeInfo: PTypeInfo; aMsg: string = '');
begin
  case aTypeInfo.Kind of
    tkEnumeration:
      AreEqualEnum(LongWord(aExpectedValue), LongWord(aActualValue), aTypeInfo, aMsg);
    tkSet:
      AreEqualSet(aExpectedValue, aActualValue, aTypeInfo, aMsg);
    tkVariant:
      AreEqualVariant(Variant(aExpectedValue), Variant(aActualValue), aMsg);
    else
      raise EDunitException.CreateFmt('aTypeInfo: %s, %s is not supported', [GetEnumName(TypeInfo(TTypeKind), Ord(aTypeInfo.Kind)), aTypeInfo.Name]);
    end;
end;

procedure TDUnitAssert.AreNotEqual(aNotExpected, aActual: Integer; const aMessage: string = '');
begin
  FTestCase.CheckNotEquals(aNotExpected, aActual, aMessage);
end;

procedure TDUnitAssert.AreNotEqual(aNotExpected, aActual: Extended; const aMessage: string = '');
begin
  FTestCase.CheckNotEquals(aNotExpected, aActual, 0, aMessage);
end;

procedure TDUnitAssert.AreNotEqual(aNotExpected, aActual: Extended; delta: Extended; const aMessage: string = '');
begin
  FTestCase.CheckNotEquals(aNotExpected, aActual, delta, aMessage);
end;

procedure TDUnitAssert.AreNotEqual(aNotExpected, aActual: Boolean; const aMessage: string = '');
begin
  FTestCase.CheckNotEquals(aNotExpected, aActual, aMessage);
end;

procedure TDUnitAssert.AreEqualDate(const aExpected, aActual: TDateTime; const aMessage: string = '');
begin
  if (abs(aExpected-aActual) > EncodeTime(0,0,0,1)) then
    FTestCase.FailNotEquals(DateTimeToStr(aExpected), DateTimeToStr(aActual), aMessage, CallerAddr)
  else
    FTestCase.AssertCalled;
end;

const
  EnumMask: array[TOrdType] of LongWord = ($FF, $FF, $FFFF, $FFFF, $FFFFFFFF, $FFFFFFFF);

procedure TDUnitAssert.AreEqualEnum(aExpectedEnum, aActualEnum: LongWord; aEnumTypeInfo: PTypeInfo; aMsg: string = '');
var
  iMask: LongWord;
begin
  iMask := EnumMask[GetTypeData(aEnumTypeInfo).OrdType];
  if (aExpectedEnum and iMask) <> (aActualEnum and iMask) then
    FTestCase.FailNotEquals(GetEnumName(aEnumTypeInfo, (aExpectedEnum and iMask)), GetEnumName(aEnumTypeInfo, (aActualEnum and iMask)), aMsg, CallerAddr)
  else
    FTestCase.AssertCalled;
end;

procedure TDUnitAssert.AreEqualSet(const aExpectedSet, aActualSet; aSetTypeInfo: PTypeInfo; aMsg: string = '');
var
  oExp: TMaxSet; // with a litle help from TypInfo.GetSetProp
  oAct: TMaxSet;
  ElementTypeData: PTypeData;
begin
  ElementTypeData := GetSetElementTypeData(aSetTypeInfo);
  ConvertSetToTMaxSet(aExpectedSet, ElementTypeData, oExp);
  ConvertSetToTMaxSet(aActualSet, ElementTypeData, oAct);
  if oExp <> oAct then
    FTestCase.FailNotEquals(SetToString(oExp, aSetTypeInfo), SetToString(oAct, aSetTypeInfo), aMsg, CallerAddr)
  else
    FTestCase.AssertCalled;
end;

// for Delphi 5 compatibility
procedure TDUnitAssert.AreEqualString(const aExpected, aActual: string; const aMessage: string = '');
begin
  FTestCase.CheckEqualsString(aExpected, aActual, aMessage);
//if aExpected <> aActual then
// FTestCase.FailNotEquals(aExpected, aActual, aMessage, CallerAddr);
end;

{$IFDEF VER130} // code missing in D5
function FindVarData(const V: Variant): PVarData;
begin
  Result := @TVarData(V);
  while Result.VType = varByRef or varVariant do
    Result := PVarData(Result.VPointer);
end;

function VarSameValue(const A, B: Variant): Boolean;
var
  LA, LB: TVarData;
begin
  LA := FindVarData(A)^;
  LB := FindVarData(B)^;
  if LA.VType = varEmpty then
    Result := LB.VType = varEmpty
  else if LA.VType = varNull then
    Result := LB.VType = varNull
  else if LB.VType in [varEmpty, varNull] then
    Result := False
  else
    Result := A = B;
end;
{$ENDIF}

// for Delphi 5 compatibility
procedure TDUnitAssert.AreNotEqualString(const aNotExpected, aActual: string; const aMessage: string = '');
begin
  FTestCase.CheckNotEqualsString(aNotExpected, aActual, aMessage);
end;

procedure TDUnitAssert.AreNotEqualSet(const aNotExpectedSet, aActualSet; aSetTypeInfo: PTypeInfo; aMsg: string = '');
var
  oNotExp: TMaxSet;
  oAct: TMaxSet;
  ElementTypeData: PTypeData;
begin
  ElementTypeData := GetSetElementTypeData(aSetTypeInfo);

  ConvertSetToTMaxSet(aNotExpectedSet, ElementTypeData, oNotExp);
  ConvertSetToTMaxSet(aActualSet, ElementTypeData, oAct);
  if oNotExp = oAct then
    FTestCase.FailEquals(SetToString(oNotExp, aSetTypeInfo), SetToString(oAct, aSetTypeInfo), aMsg, CallerAddr)
  else
    FTestCase.AssertCalled;
end;

procedure TDUnitAssert.AreNotEqualEnum(aNotExpectedEnum, aActualEnum: LongWord; aEnumTypeInfo: PTypeInfo; aMsg: string
    = '');
var
  iMask: LongWord;
begin
  iMask := EnumMask[GetTypeData(aEnumTypeInfo).OrdType];
  if (aNotExpectedEnum and iMask) = (aActualEnum and iMask) then
    FTestCase.FailEquals(GetEnumName(aEnumTypeInfo, (aNotExpectedEnum and iMask)), GetEnumName(aEnumTypeInfo, (aActualEnum and iMask)), aMsg, CallerAddr)
  else
    FTestCase.AssertCalled;
end;

procedure TDUnitAssert.AreNotEqual(const aExpectedValue, aActualValue; aTypeInfo: PTypeInfo; aMsg: string = '');
begin
  case aTypeInfo.Kind of
    tkEnumeration:
      AreNotEqualEnum(LongWord(aExpectedValue), LongWord(aActualValue), aTypeInfo, aMsg);
    tkSet:
      AreNotEqualSet(aExpectedValue, aActualValue, aTypeInfo, aMsg);
    tkVariant:
      AreNotEqualVariant(Variant(aExpectedValue), Variant(aActualValue), aMsg);
    else
      raise EDunitException.CreateFmt('aTypeInfo: %s, %s is not supported', [GetEnumName(TypeInfo(TTypeKind), Ord(aTypeInfo.Kind)), aTypeInfo.Name]);
    end;
end;

procedure TDUnitAssert.AreEqualVariant(const aExpected, aActual: Variant; aMsg: string = '');
begin
  if not VarSameValue(aExpected, aActual) then
    TestCase.FailNotEquals(VarToStr(aExpected), VarToStr(aActual), aMsg, CallerAddr)
  else
    TestCase.AssertCalled;
end;

procedure TDUnitAssert.AreNotEqualVariant(const aNotExpected, aActual: Variant; aMsg: string = '');
begin
  if VarSameValue(aNotExpected, aActual) then
    TestCase.FailEquals(VarToStr(aNotExpected), VarToStr(aActual), aMsg, CallerAddr)
  else
    TestCase.AssertCalled;
end;

// for Delphi 5 compatibility
procedure TDUnitAssert.AreEqualWideString(const aExpected, aActual: WideString; const aMessage: string = '');
begin
  FTestCase.CheckEqualsWideString(aExpected, aActual, aMessage);
//  if aExpected <> aActual then
//     FTestCase.FailNotEquals(aExpected, aActual, aMessage, CallerAddr);
end;

// for Delphi 5 compatibility
procedure TDUnitAssert.AreNotEqualWideString(const aNotExpected, aActual: WideString; const aMessage: string = '');
begin
  FTestCase.CheckNotEqualsWideString(aNotExpected, aActual, aMessage);
end;

procedure TDUnitAssert.AreNotEqual(aNotExpected, aActual: TClass; const aMessage: string = '');
begin
 if aNotExpected = aActual then
 begin
   if aNotExpected = nil then
     FTestCase.FailEquals('nil', 'nil', aMessage, CallerAddr)
   else
     FTestCase.FailEquals(aNotExpected.ClassName, aActual.ClassName, aMessage, CallerAddr)
 end
 else
   FTestCase.AssertCalled;
end;

procedure TDUnitAssert.AreNotEqual(const aNotExpected, aActual: string; const aMessage: string = '');
begin
  {$IFNDEF VER130}
  FTestCase.CheckNotEquals(aNotExpected, aActual, aMessage);
  {$ELSE} // D5
  AreNotEqualString(aNotExpected, aActual, aMessage);
  {$ENDIF}
end;

procedure TDUnitAssert.AreNotEqual(const aNotExpected, aActual: WideString; const aMessage: string = '');
begin
  {$IFNDEF VER130}
  FTestCase.CheckNotEquals(aNotExpected, aActual, aMessage);
  {$ELSE} // D5
  AreNotEqualWideString(aNotExpected, aActual, aMessage);
  {$ENDIF}
end;

procedure TDUnitAssert.AreNotEqual(const aNotExpected, aActual: TDateTime; const aMessage: string = '');
begin
  AreNotEqualDate(aNotExpected, aActual, aMessage);
end;

procedure TDUnitAssert.AreNotEqualDate(const aNotExpected, aActual: TDateTime; const aMessage: string = '');
begin
  if (abs(aNotExpected-aActual) < EncodeTime(0,0,0,1)) then
    FTestCase.FailEquals(DateTimeToStr(aNotExpected), DateTimeToStr(aActual), aMessage, CallerAddr)
  else
    FTestCase.AssertCalled;
end;

procedure TDUnitAssert.AreSame(aExpected, aActual: IInterface; const aMessage: string = '');
begin
  FTestCase.CheckSame(aExpected, aActual, aMessage);
end;

procedure TDUnitAssert.AreSame(aExpected, aActual: TObject; const aMessage: string = '');
begin
  FTestCase.CheckSame(aExpected, aActual, aMessage);
end;

procedure TDUnitAssert.AreNotSame(aNotExpected, aActual: TObject; const aMessage: string = '');
begin
  if (aNotExpected = aActual) then
    FTestCase.Fail(FTestCase.EqualsErrorMessage(PtrToStr(Pointer(aNotExpected)), PtrToStr(Pointer(aActual)), aMessage), CallerAddr)
  else
    FTestCase.AssertCalled;
end;

procedure TDUnitAssert.AreNotSame(aNotExpected, aActual: IInterface; const aMessage: string = '');
begin
  if (aNotExpected = aActual) then
    FTestCase.Fail(FTestCase.EqualsErrorMessage(PtrToStr(Pointer(aNotExpected)), PtrToStr(Pointer(aActual)), aMessage), CallerAddr)
  else
    FTestCase.AssertCalled;
end;

procedure TDUnitAssert.IsInstanceOfType(const aExpectedClass: TClass; const aObject: TObject; const aMessage: string = '');
begin
  FTestCase.CheckIs(aObject, aExpectedClass, aMessage);
end;

procedure TDUnitAssert.Fail(const aMessage: string);
begin
  FTestCase.Fail(aMessage);
end;

class function TDUnitAssert.GetSetElementTypeData(aSetTypeInfo: PTypeInfo): PTypeData;
var
  SetTypeData: PTypeData;
  ElementTypeInfo: PTypeInfo;
begin
  if aSetTypeInfo.Kind <> tkSet then
    raise EDunitException.Create('aSetTypeInfo should be of tkSet type and not of '+GetEnumName(Typeinfo(TTypeKind), Ord(aSetTypeInfo.Kind)));
  SetTypeData := GetTypeData(aSetTypeInfo);
  if SetTypeData.CompType=nil then
    raise EDunitException.CreateFmt('The set %s has no typeinfo for it''s element. Are you using enums with numeric values?', [aSetTypeInfo.Name]);

  ElementTypeInfo := SetTypeData.CompType^;
  Result := GetTypeData(ElementTypeInfo);
end;

class procedure TDUnitAssert.ConvertSetToTMaxSet(const aSet; aElementTypeData: PTypeData; out aMaxSet:TMaxSet);
begin
  aMaxSet := [];
  Move(aSet, aMaxSet, 1+(aElementTypeData.MaxValue div 8));
end;

procedure TDUnitAssert.IsTrue(aCondition: Boolean; const aMessage: string='');
begin
  FTestCase.CheckTrue(aCondition, aMessage)
end;

procedure TDUnitAssert.IsFalse(aCondition: Boolean; const aMessage: string='');
begin
  FTestCase.CheckFalse(aCondition, aMessage)
end;

procedure TDUnitAssert.IsNotInstanceOfType(const aExpectedClass: TClass; const aObject: TObject; const aMessage: string
    = '');
begin
  if (AObject <> nil) and AObject.ClassType.InheritsFrom(aExpectedClass) then
    FTestCase.FailNotEquals(aExpectedClass.ClassName, AObject.ClassName, aMessage, CallerAddr)
  else
    FTestCase.AssertCalled;
end;

procedure TDUnitAssert.IsNotNull(aObj: TObject; const aMessage: string='');
begin
    FTestCase.CheckNotNull(aObj, aMessage);
end;

procedure TDUnitAssert.IsNotNull(aObj: IInterface; const aMessage: string='');
begin
  FTestCase.CheckNotNull(aObj, FTestCase.NotSameErrorMessage('interface', PtrToStr(Pointer(aObj)), aMessage));
end;

procedure TDUnitAssert.IsNull(aObj: TObject; const aMessage: string='');
begin
  FTestCase.CheckNull(aObj, aMessage);
end;

procedure TDUnitAssert.IsNull(aObj: IInterface; const aMessage: string='');
begin
  FTestCase.CheckNull(aObj, aMessage);
end;

class function TDUnitAssert.SetToString(const aSetValue; aSetTypeInfo: PTypeInfo): string;
var
  S: TMaxSet;
  ElementTypeInfo: PTypeInfo;
  oTypeData: PTypeData;
  I: Integer;
  ElementTypeData: PTypeData;
begin
  if aSetTypeInfo.Kind <> tkSet then
    raise EDunitException.Create('SetToString works only on sets');
  S := TMaxSet(aSetValue);
  oTypeData := GetTypeData(aSetTypeInfo);
  ElementTypeInfo := oTypeData.CompType^;
  ElementTypeData := GetTypeData(ElementTypeInfo);
  Result := '';
  for I := ElementTypeData.MinValue to ElementTypeData.MaxValue do
    if I in S then
    begin
      if Result <> '' then
        Result := Result + ',';
      case ElementTypeInfo.Kind of
        tkChar: Result := Result + Format('''%s''(#%d)', [Char(I), I]);
        tkInteger,
        tkEnumeration: Result := Result + GetEnumName(ElementTypeInfo, I);
      end;
    end;
  Result := '[' + Result + ']';
end;

{ TAssertTestCase }

procedure TAssertTestCase.AssertCalled;
begin
  FCheckCalled := True;
end;

constructor TAssertTestCase.Create(MethodName: string);
begin
  inherited;
  FAssert := CreateAssert;
end;

destructor TAssertTestCase.Destroy;
begin
  FreeAndNil(FAssert);
  inherited;
end;

function TAssertTestCase.CreateAssert: TDUnitAssert;
begin
  Result := TDUnitAssert.Create(Self);
end;


initialization
// calc in AreEqualSet assume the following
  Assert(SizeOf([0..7])=1);
  Assert(SizeOf([0..15])=2);
  Assert(SizeOf([0..247])=31);
  Assert(SizeOf([0..255])=32);
end.

